import unittest

import matplotlib.pyplot as plt
import numpy as np

from tart.simulation.simulation_source import SimulationSource
from tart.util import angle


class TestSimulationSource(unittest.TestCase):
    def setUp(self):
        self.crab = SimulationSource(
            r=1e9,
            amplitude=1.0,
            azimuth=angle.from_dms(0.0),
            elevation=angle.from_dms(45.0),
            sample_duration=1e-3,
        )  # 0.1 ms sample duration
        self.s2 = SimulationSource(
            r=1e9,
            amplitude=1.0,
            azimuth=angle.from_dms(0.0),
            elevation=angle.from_dms(90.0),
            sample_duration=1e-3,
        )  # 0.1 ms sample duration
        self.s3 = SimulationSource(
            r=1e9,
            amplitude=1.0,
            azimuth=angle.from_dms(0.0),
            elevation=angle.from_dms(0.0),
            sample_duration=1e-3,
        )  # 0.1 ms sample duration

    def test_full_simp(self):
        tend = 1e-4
        full_time = np.arange(0, tend, 1.0 / (1.575e9))
        crab_s = self.crab.s(full_time)

        samp_time = np.arange(0, tend, 1.0 / (16.0e6))
        samp_time2 = np.arange(0, tend, 1.0 / (16.0e6)) + 1e-6

        crab_s_baseband = self.crab.s_baseband(samp_time)
        crab_s_baseband2 = self.crab.s_baseband(samp_time2)
        plt.plot(full_time, crab_s / np.cos(full_time * self.crab.omega), label="full")
        plt.plot(samp_time, crab_s_baseband, label="sampled")
        plt.plot(samp_time2, crab_s_baseband2, label="sampled")

        plt.legend()
        plt.show()

    def test_signal_deterministic(self):
        samp_freq = 16.0e6
        samp_time = np.arange(0, 1e-4, 1.0 / (samp_freq))
        a = self.crab.s_baseband(samp_time)
        dt = 1.0 / (16.0e6)
        b = self.crab.s_baseband(samp_time + dt)

        for i in np.abs(a[1:] - b[:-1]):
            self.assertLess(i, 1e-5)

        dt = 0.5 / (16.0e6)
        b = self.crab.s_baseband(samp_time + dt)
        for i in range(1, 100):
            print(i)
            if a[i + 1] - a[i] > 0:
                self.assertTrue(a[i + 1] - a[i] > a[i + 1] - b[i])
                self.assertTrue(a[i + 1] - a[i] > b[i] - a[i])

    def test_signal(self):
        timebase = np.arange(0, 1e-3, 2e-10)
        crab_signal = self.crab.s(timebase)
        mean = crab_signal.mean()
        stdev = crab_signal.std()
        self.assertLess(abs(mean), 1.0e-6)
        self.assertLess(stdev, 0.40)  # Should be 0.33
        self.assertGreater(stdev, 0.31)

    def test_baseband(self):
        timebase = np.arange(0, 1e-3, 1e-7)
        crab_signal = self.crab.s_baseband(timebase).real
        mean = crab_signal.mean()
        stdev = crab_signal.std()
        self.assertLess(abs(mean), 2.0e-2)
        self.assertLess(stdev, 0.55)  # Should be 0.5
        self.assertGreater(stdev, 0.47)



import pickle

import numpy as np

# import pyfftw.interfaces.numpy_fft as fft
from numpy import fft

from tart.imaging import location, radio_source
from tart.simulation import antennas
from tart.util import angle, constants

cc = np.concatenate


def get_max_ang(nw, num_bin):
    ret = np.degrees(num_bin / (4.0 * nw))
    return ret


class Synthesis_Imaging:
    def __init__(self, cal_vis_list):
        self.cal_vis_list = cal_vis_list
        self.phase_center = None
        self.grid_file = "grid.idx"
        self.grid_idx = None

    def set_grid_file(self, fpath):
        self.grid_file = fpath

    def get_uuvvwwvis_zenith(self):
        vis_l = []
        # for cal_vis in copy.deepcopy(self.cal_vis_list[:1]):
        for cal_vis in self.cal_vis_list[:1]:
            ant_p = np.array(cal_vis.get_config().get_antenna_positions())
            bls = cal_vis.get_baselines()
            pos_pairs = ant_p[np.array(bls)]
            uu_a, vv_a, ww_a = (
                pos_pairs[:, 0] - pos_pairs[:, 1]
            ).T / constants.L1_WAVELENGTH
            for bl in bls:
                ant_i, ant_j = bl
                vis_l.append(cal_vis.get_visibility(ant_i, ant_j))
        return uu_a, vv_a, ww_a, np.array(vis_l)

    def get_grid_idxs(self, uu_a, vv_a, num_bin, nw):
        try:
            if self.grid_idx is None:
                self.grid_idx = pickle.load(open(self.grid_file, "rb"))
                # print('finished loading ' + self.grid_file)
        except:
            print("generating...")
            uu_edges = np.linspace(-nw, nw, num_bin + 1)
            vv_edges = np.linspace(-nw, nw, num_bin + 1)
            grid_idx = []
            for uu, vv in zip(uu_a, vv_a):
                i = uu_edges.__lt__(uu).sum() - 1
                j = vv_edges.__lt__(vv).sum() - 1
                i2 = uu_edges.__lt__(-uu).sum() - 1
                j2 = vv_edges.__lt__(-vv).sum() - 1
                grid_idx.append([i, j, i2, j2])
            self.grid_idx = np.array(grid_idx)
            save_ptr = open(self.grid_file, "wb")
            pickle.dump(self.grid_idx, save_ptr, pickle.HIGHEST_PROTOCOL)
            save_ptr.close()
        return self.grid_idx

    def get_uvplane_zenith(self, num_bin=1600, nw=36):
        uu_a, vv_a, ww_a, vis_l = self.get_uuvvwwvis_zenith()
        arr = np.zeros((num_bin, num_bin), dtype=np.complex64)
        # place complex visibilities in the UV grid and prepare averaging by counting entries.
        grid_idxs = self.get_grid_idxs(uu_a, vv_a, num_bin, nw)
        count_arr = np.zeros((num_bin, num_bin), dtype=np.int16)

        arr[grid_idxs[:, 1], grid_idxs[:, 0]] = vis_l
        arr[grid_idxs[:, 3], grid_idxs[:, 2]] = np.conjugate(vis_l)
        n_arr = np.ma.masked_array(arr[:, :], count_arr.__lt__(1.0))
        return n_arr

    def get_uvplane(self, num_bin=1600, nw=36, grid_kernel_r_pixels=0.5):
        for cal_vis in self.cal_vis_list:
            uu_a, vv_a, ww_a = cal_vis.get_all_uvw()
            uu_a = uu_a / constants.L1_WAVELENGTH
            vv_a = vv_a / constants.L1_WAVELENGTH
            ww_a = ww_a / constants.L1_WAVELENGTH
            vis_l, bls = cal_vis.get_all_visibility()

        uu_edges = np.linspace(-nw, nw, num_bin + 1)
        vv_edges = np.linspace(-nw, nw, num_bin + 1)

        # TODO: Throw an exception if the UV-plant does not have sufficient
        # resolution to manage the baselines.

        n_arr = np.zeros((num_bin, num_bin), dtype=np.complex64)
        uu_comb = np.concatenate((uu_a, -uu_a))
        vv_comb = np.concatenate((vv_a, -vv_a))
        all_v = np.concatenate((vis_l, np.conjugate(vis_l)))
        h_real, _, _ = np.histogram2d(
            vv_comb, uu_comb, weights=all_v.real, bins=[vv_edges, uu_edges]
        )
        h_imag, _, _ = np.histogram2d(
            vv_comb, uu_comb, weights=all_v.imag, bins=[vv_edges, uu_edges]
        )
        num_entries, _, _ = np.histogram2d(
            vv_comb, uu_comb, bins=[vv_edges, uu_edges]
        )
        n_arr[:, :] = h_real + (1j * h_imag)
        pos = np.where(num_entries.__gt__(1))
        n_arr[pos] /= num_entries[pos]

        return (n_arr, uu_edges, vv_edges)

    def get_ift(self, nw=30, num_bin=2 ** 7):
        uv_plane, uu_edges, vv_edges = self.get_uvplane(
            num_bin=num_bin, nw=nw)
        ift = np.fft.fftshift(fft.ifft2(np.fft.ifftshift(uv_plane)))
        maxang = get_max_ang(nw, num_bin)
        extent = [maxang, -maxang, -maxang, maxang]
        return [ift, extent]

    def get_beam(self, nw=30, num_bin=2 ** 7):
        uv_plane, uu_edges, vv_edges = self.get_uvplane(
            num_bin=num_bin, nw=nw)
        ift = np.fft.ifftshift(fft.ifft2(np.fft.ifftshift(np.abs(uv_plane).__gt__(0))))
        return ift  # /np.sum(ret)

    def get_image(self, CAL_IFT, CAL_EXTENT):
        abs_ift = np.abs(CAL_IFT)
        ift_std = np.std(abs_ift)
        ift_scaled = abs_ift / ift_std

        plt.figure(figsize=(8, 6))
        plt.imshow(ift_scaled, extent=CAL_EXTENT)
        plt.colorbar()
        ift_median = np.median(abs_ift)
        ts = self.cal_vis_list[0].get_timestamp()
        plt.title(
            ts.strftime("%d-%m-%y %H:%M:%S")
            + " std %1.3e median: %1.3e" % (ift_std, ift_median)
        )
        plt.grid()
        plt.xlim(63, -63)
        plt.ylim(-63, 63)
        plt.xlabel("East-West")
        plt.ylabel("North-South")
        plt.tight_layout()



# from tart.imaging import radio_source
# from tart.imaging.ephemerides_proxy import EphemeridesProxy
# from tart.util import utc
# from tart.util import constants
# from tart.util import vector
import unittest

from tart.imaging import location
from tart.imaging.gps_satellite import *
from tart.util import angle, utc


class TestGpsSatellite(unittest.TestCase):
    def setUp(self):
        self.t = utc.utc_datetime(2002, 10, 31, 0, 2, 2)

    def test_gps(self):

        import ephem

        dnd = ephem.Observer()
        dnd.lon = str(location.Dunedin.lon.to_degrees())
        dnd.lat = str(location.Dunedin.lat.to_degrees())
        dnd.elevation = location.Dunedin.alt
        dnd.pressure = 0.0
        t = utc.utc_datetime(2013, 9, 20, 0, 0, 3)
        dnd.date = t

        for svnum in range(31, 32):
            l1 = "GPS BIIRM-2 (PRN 31)    "
            l2 = "1 29486U 06042A   13264.26023969 -.00000084  00000-0  00000+0 0  2292"
            l3 = "2 29486  56.1466 336.5502 0081195 316.5190  69.5394  2.00563866 51246"
            j = ephem.readtle(l1, l2, l3)
            # http://blog.thetelegraphic.com/2012/gps-sattelite-tracking-in-python-using-pyephem/
            # j = ephem.Gps(svnum)
            j.compute(dnd)
            ra_e = angle.from_rad(j.ra.real)
            dec_e = angle.from_rad(j.dec.real)

            sv = GpsSatellite(svnum, location.Dunedin)
            ra, dec = sv.radec(t)

            print(angle.wrap_360(ra.to_degrees()))
            self.assertAlmostEqual(ra.to_degrees(), ra_e.to_degrees(), 0)
            self.assertAlmostEqual(dec.to_degrees(), dec_e.to_degrees(), 0)

    def test_radec(self):
        sv = GpsSatellite(31, location.Dunedin)
        t = utc.utc_datetime(2013, 9, 20, 0, 0, 3)
        ra, dec = sv.radec(t)
        self.assertLess(ra.to_degrees(), 360.0)
        self.assertLess(abs(dec.to_degrees()), 90.0)

    def test_power(self):
        sv = GpsSatellite(31, location.Dunedin)
        t = utc.utc_datetime(2013, 9, 20, 0, 0, 3)
        self.assertLess(sv.jansky(t), 5.0e6)
        self.assertGreater(sv.jansky(t), 3.0e6)



"""Models of Antennas"""

import numpy as np

from tart.imaging import visibility
from tart.imaging import imaging
from tart.simulation.simulation_source import HorizontalSource
from tart.util import constants


def antennas_signal(antennas, ant_models, sources, timebase):
    """ TODO think about introducing uncorrelated system noise at each antenna."""
    ret = np.zeros((len(antennas), len(timebase)))
    # An array to hold the signal seen by each antenna (in rows)
    # One row for each antenna

    for i, ant in enumerate(antennas):
        # A vector to hold the working value as signals are added
        working = np.zeros(len(timebase))
        for src in sources:
            delta = ant.get_geo_delay_horizontal(src)
            # print("Delay %s: %g" % (ant, dt))
            gain = ant_models[i].get_gain(src.elevation, src.azimuth)
            # print('Antenna: %i Gain: %1.1f el: %1.1f az: %1.1f' % (i, gain, src.elevation.to_degrees(), src.azimuth.to_degrees()))
            working += src.s(timebase + delta) * gain
        ret[i, :] = working
    return ret


def antennas_simplified_signal(antennas, ant_models, sources, timebase, fc0, seed=None):
    """ Retun array of simplified (at baseband) antenna signals ."""
    np.random.seed(seed=seed)
    debug = False
    # create an array to hold the signal seen by each antenna (in rows)
    ant_sigs = []

    for i, ant in enumerate(antennas):
        s_bb = np.zeros(len(timebase), dtype=complex)
        for src in sources:  # Cycle through each signal source in turn
            gain = ant_models[i].get_gain(src.elevation, src.azimuth)
            if gain > 0.0:  # Gain theshold
                dt = ant.get_geo_delay_horizontal(src)
                s_bb += (
                    src.s_baseband(timebase + dt) * gain * np.exp(1j * src.omega * dt)
                )
        ant_sigs.append(s_bb)
    return np.asarray(ant_sigs)  # signal in baseband


def antennas_simp_vis(antennas, ant_models, sources, utc_date, config, noise_lvl):
    """ Return visibility object without generating timeseries or filtering."""
    vis = []
    num_ant = len(antennas)
    baseline_indices = imaging.get_baseline_indices(num_ant)

    # noise = np.random.uniform(0.,np.sqrt(noise_lvl),config.get_num_antenna()) * np.exp(2.0j*np.pi*np.random.uniform(-1.,1.,config.get_num_antenna()))
    if noise_lvl.__gt__(0.0).all():
        noise = np.random.normal(0.0, noise_lvl) * np.exp(
            2.0j * np.pi * np.random.uniform(-1.0, 1.0, num_ant)
        )
    else:
        noise = np.zeros(num_ant)

    for bl_idx in baseline_indices:
        i, j = bl_idx
        vi = noise[i] + noise[j]

        for src in sources:
            gain0 = ant_models[i].get_gain(src.elevation, src.azimuth)
            gain1 = ant_models[j].get_gain(src.elevation, src.azimuth)
            if gain0 <= 0.0 or gain1 <= 0.0:
                vi += 0.0j
            else:
                dt = get_geo_delay_horizontal(antennas[i], antennas[j], src)
                vi += (
                    gain0
                    * gain1
                    * 1.0
                    * np.exp(1.0j * dt * constants.L1_OMEGA)
                    * src.amplitude
                )
        if np.abs(vi) >= 1.0:
            # otherwise in case we have a signal that is almost 1. noise could cause an overflow
            vi = 1.0 * vi / np.abs(vi)
        vis.append(vi)

    vis_o = visibility.Visibility.from_config(config, utc_date)
    vis_o.set_visibilities(vis, baseline_indices)
    return vis_o


class Antenna:
    """Holds position of the Antenna and can calculate the geometric
    delay for given horizontal (el, az) or equatorial (time, RA, DEC) coordinates"""

    def __init__(self, loc, enu):
        self.loc = loc
        self.enu = enu
        self.dxdydz = self.loc.get_ecef_delta_from_enu(enu[0], enu[1], enu[2])

    def __repr__(self):
        return f"Antenna: enu={self.enu}"

    def get_geo_delay(self, sloc, utc_time):
        """
        Calculate and return geometric delay for each antenna
        Negative value means the wavefront hits the antenna earlier (than the zero ENU location)
        """
        el, az = self.loc.equatorial_to_horizontal(utc_time, sloc.ra, sloc.dec)
        return self.get_geo_delay_horizontal(el, az)

    def get_geo_delay_horizontal(self, src_0):
        """
        Calculate and return geometric delay for each antenna
        Negative value means the wavefront hits the antenna earlier (than the zero ENU location)
        t_antenna - t_origin
        """
        el_0, az_0, r_0 = src_0.elevation, src_0.azimuth, src_0.r
        if src_0.r > 1e4:
            src_0.r = 1.0e4

        # el_0 and az_0 from the PoV of the antenna, corrected for the distance of the source.
        # if source is far away: el = el_0 and az = az_0
        object_vector = np.zeros(3)
        object_vector[0] = az_0.sin() * el_0.cos()
        object_vector[1] = az_0.cos() * el_0.cos()
        object_vector[2] = el_0.sin()

        object_vector_ant_pov = r_0 * object_vector - self.enu

        r = np.linalg.norm(object_vector_ant_pov)
        path_diff = r - r_0
        delay = path_diff / constants.V_LIGHT
        # print(delay)
        return delay

    def calcUVW(self, utc_time, ra, dec):
        """u is towards the east, v is towards the north,
        w points to the phase centre"""

        x, y, z = self.dxdydz
        ha = self.loc.GHA(utc_time, ra)
        u = ha.sin() * x + ha.cos() * y
        v = -dec.sin() * ha.cos() * x + dec.sin() * ha.sin() * y + dec.cos() * z
        w = dec.cos() * ha.cos() * x - dec.cos() * ha.sin() * y + dec.sin() * z
        return np.array([u, v, w])


def get_geo_delay_horizontal(a0, a1, src):
    assert isinstance(src, HorizontalSource)
    """
    The delay is negative if a1 is closer to the source than a0

    t_a1 - t_a0 = (t_a1 - t_origin) - (t_a0 - t_origin)
    """
    d0 = a0.get_geo_delay_horizontal(src)
    d1 = a1.get_geo_delay_horizontal(src)
    return d1 - d0


def get_UVW(a0, a1, utc_time, ra, dec):
    uvw0 = a0.calcUVW(utc_time, ra, dec)
    uvw1 = a1.calcUVW(utc_time, ra, dec)
    # print(uvw0, uvw1, uvw1 - uvw0)
    return uvw0 - uvw1



import unittest
import logging
import os

import numpy as np
from copy import deepcopy

from tart.util import utc

from tart.imaging import elaz
from tart.imaging import imaging
from tart.operation import settings

logger = logging.getLogger()

TESTCONFIG_FILENAME = os.path.join(os.path.dirname(__file__), '../../test/test_telescope_config.json')
ANT_POS_FILE = os.path.join(os.path.dirname(__file__), '../../test/test_calibrated_antenna_positions.json')


class TestImaging(unittest.TestCase):

    def setUp(self):
        self.config = settings.from_file(TESTCONFIG_FILENAME)
        self.config.load_antenna_positions(
            cal_ant_positions_file=ANT_POS_FILE
        )

    def test_clock_sources(self):
        hour_sources, minute_sources = imaging.get_clock_hands(timestamp=utc.now())

        n_bin = 2 ** 6
        img = np.zeros((n_bin, n_bin))
        img[n_bin-1, 0] = 1
        import matplotlib.pyplot as plt

        for m in minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            img[x, y] = 1

        for m in hour_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            img[x, y] = 1

        plt.imshow(img)
        plt.show()

        for m in minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            self.assertGreater(img[x, y], 0.5)

    def test_imaging(self):

        cv, hour_sources, minute_sources = \
            imaging.get_clock_vis(timestamp=utc.now(), config=self.config)

        n_bin = 2 ** 8

        cal_ift, cal_extent, n_fft, bin_width = imaging.image_from_calibrated_vis(
            cv, nw=n_bin / 4, num_bin=n_bin
        )

        img = np.abs(cal_ift)
        max_p = np.max(img)
        min_p = np.min(img)
        mad_p = np.std(img)
        ift_scaled = (img) / mad_p

        import matplotlib.pyplot as plt

        for m in minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            print(x, y)
            ift_scaled[x, y] = -1

        plt.imshow(ift_scaled)
        plt.show()

        for m in minute_sources:
            src = elaz.ElAz(m['el'], m['az'])
            x, y = src.get_px(n_bin)
            self.assertGreater(img[x, y], max_p/3)

    def test_lm_to_index(self):
        image_size = 256

        # Center
        i0, i1 = imaging.get_lm_index(0.0, 0.0, image_size)
        self.assertEqual(i0, image_size // 2)
        self.assertEqual(i1, image_size // 2)


        # North (top center)
        i0, i1 = imaging.get_lm_index(0.0, 1.0, image_size)
        self.assertEqual(i0, 0)
        self.assertEqual(i1, image_size // 2)

        # West (left center)
        i0, i1 = imaging.get_lm_index(-1.0, 0.0, image_size)
        self.assertEqual(i0, image_size // 2)
        self.assertEqual(i1, 0)

        # East (right center)
        i0, i1 = imaging.get_lm_index(1.0, 0.0, image_size)
        self.assertEqual(i0, image_size // 2)
        self.assertEqual(i1, image_size - 1)

        # South (bottom center)
        i0, i1 = imaging.get_lm_index(0.0, -1.0, image_size)
        self.assertEqual(i0, image_size - 1)
        self.assertEqual(i1, image_size // 2)


        # Upper right corner
        i0, i1 = imaging.get_lm_index(1.0, 1.0, image_size)
        self.assertEqual(i0, 0)
        self.assertEqual(i1, image_size-1)

        # Upper left corner
        i0, i1 = imaging.get_lm_index(-1.0, 1.0, image_size)
        self.assertEqual(i0, 0)
        self.assertEqual(i1, 0)

        # Lower right corner
        i0, i1 = imaging.get_lm_index(1.0, -1.0, image_size)
        self.assertEqual(i0, image_size-1)
        self.assertEqual(i1, image_size-1)

        # Lower left corner
        i0, i1 = imaging.get_lm_index(-1.0, -1.0, image_size)
        self.assertEqual(i0, image_size-1)
        self.assertEqual(i1, 0)







#
# A class the encapsulates a location on earth. Does all the necessary frame-of-reference
# conversions.
#
# Copyright (c) Tim Molteno 2013-2017. tim@elec.ac.nz
#

import math

import numpy as np

from tart.imaging import tart_util
from tart.util import angle


def get_loc(Settings):
    return Location(
        angle.from_dms(Settings.get_lat()),
        angle.from_dms(Settings.get_lon()),
        Settings.get_alt(),
    )


""" Future class structure going in here"""


class ENU:
    def __init__(self, e, n, u):
        self.e = e
        self.n = n
        self.u = u


class Horizontal:
    def __init__(self, r, el, az):
        self.r = r
        self.el = el
        self.az = az


class ECEF:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z


class ECI:
    def __init__(self, x, y, z):
        self.x = x
        self.y = y
        self.z = z


# Convert ECI (Earth-Centered-Inertial) to ECEF coordinates.
def eci_to_ecef(utc_date, x_in, y_in, z_in):
    """cos(theta) -sin(theta)  0
    sin(theta)  cos(theta)  0
    0           0           1
    """
    # Rotate x and y by the hour angle
    theta = -Location.GST(utc_date)

    return [
        x_in * theta.cos() - y_in * theta.sin(),
        y_in * theta.cos() + x_in * theta.sin(),
        z_in,
    ]


# Convert ECEF to ECI (Earth-Centered-Inertial)
def ecef_to_eci(utc_date, x_in, y_in, z_in):

    # Undo Rotate x and y by the hour angle
    theta = -Location.GST(utc_date)

    return [
        x_in * theta.cos() + y_in * theta.sin(),
        y_in * theta.cos() - x_in * theta.sin(),
        z_in,
    ]


class Location:
    R_EARTH = 6378137.0  # earth semimajor axis in meters
    F_RECIP = 1.0 / 298.257223563  # reciprocal flattening
    E2 = 2.0 * F_RECIP - F_RECIP * F_RECIP  # eccentricity squared

    def __init__(self, lat, lon, alt):
        self.lat = lat  # requires init with lat and lon as angle objects (NEW)
        self.lon = lon  # requires init with lat and lon as angle objects (NEW)
        self.alt = alt
        self.ex, self.ey, self.ez = self.calculate_ecef()

    def __repr__(self):
        return "[lat=%f, lon=%f, alt=%f]" % (
            self.latitude_deg(),
            self.longitude_deg(),
            self.alt,
        )

    def longitude_deg(self):
        return angle.wrap_180(self.lon.to_degrees())

    def latitude_deg(self):
        return self.lat.to_degrees()

    def altitude(self):
        return self.alt

    def get_XYZ(self, e, n, u):
        """Return the location in terms of the XYZ (non-earth fixed coordinate system).
        The X axis starts at the earths center and points through (HA = 0, decl = 0)
        The Y axis points at 90 degrees to X and in the plane of the equator (HA = -90, decl=0)
        Z points through the north celestial pole (decl = 90)"""
        ecef0 = self.get_ecef()  # ECEF of origin
        # Now create a new location offset by ENU
        d_lon = angle.from_rad((e / self.R_EARTH) * self.lat.sin())
        d_lat = angle.from_rad(n / self.R_EARTH)

        loc2 = Location((self.lat + d_lat), (self.lon + d_lon), self.alt + u)
        ecef2 = loc2.get_ecef()

        return np.array(ecef2) - np.array(ecef0)

    def get_ecef_wgs84(self):
        a = 6378137.0  # m Semi-major axis
        b = 6356752.314245  # Semi-minor axis
        inverse_flattening = 298.257223563

        sinlat = self.lat.sin()
        n = a * a / (np.sqrt(a * a * self.lat.cos() ** 2 + b * b * sinlat ** 2))
        x = (n + self.alt) * self.lat.cos() * self.lon.cos()
        y = (n + self.alt) * self.lat.cos() * self.lon.sin()
        z = (n * (b ** 2 / a ** 2) + self.alt) * sinlat
        return [x, y, z]

    def calculate_ecef(self):
        """ Position of the Location in ECEF coordinates """
        sinlat = self.lat.sin()
        chi = math.sqrt(1.0 - Location.E2 * sinlat * sinlat)
        coslat = self.lat.cos()
        achih = Location.R_EARTH / chi + self.alt
        x = achih * coslat * self.lon.cos()
        y = achih * coslat * self.lon.sin()
        z = (Location.R_EARTH * (1.0 - Location.E2) / chi + self.alt) * sinlat

        return [x, y, z]

    def get_ecef(self):
        return [self.ex, self.ey, self.ez]

    # ECEF to ENU
    def ecef_to_enu(self, x_in, y_in, z_in):
        lon = self.lon
        lat = self.lat
        e = -x_in * lon.sin() + y_in * lon.cos()
        n = (
            -x_in * lon.cos() * lat.sin()
            - y_in * lon.sin() * lat.sin()
            + z_in * lat.cos()
        )
        u = (
            x_in * lon.cos() * lat.cos()
            + y_in * lon.sin() * lat.cos()
            + z_in * lat.sin()
        )
        return [e, n, u]

    def get_ecef_delta_from_enu(self, e_in, n_in, u_in):
        # FIXME: definitely TEST ENU_TO_ECEF
        lat = self.lat
        lon = self.lon
        x = (
            -lon.sin() * e_in
            - lon.cos() * lat.sin() * n_in
            + lon.cos() * lat.cos() * u_in
        )
        y = (
            lon.cos() * e_in
            - lon.sin() * lat.sin() * n_in
            + lon.sin() * lat.cos() * u_in
        )
        z = lat.cos() * n_in + lat.sin() * u_in
        return [x, y, z]

    # http://www.navipedia.net/index.php/Transformations_between_ECEF_and_ENU_coordinates
    def ecef_to_horizontal(self, x_in, y_in, z_in):

        ex, ey, ez = self.get_ecef()  # My position in ECEF

        rx, ry, rz = [x_in - ex, y_in - ey, z_in - ez]
        enu = np.array(self.ecef_to_enu(rx, ry, rz))

        r = np.sqrt(enu.dot(enu))
        rho = enu / r

        el = angle.asin(rho[2])

        n = rho[1]  # n
        e = rho[0]  # e

        az = angle.atan2(e, n)
        if az.to_degrees() < 0.0:
            az = angle.from_dms(360.0 + az.to_degrees())

        return [r, el, az]

    @staticmethod
    def horizontal_to_enu(r, el, az):
        """Get ENU from Horizontal (local Topocentric)"""
        e = el.cos() * az.sin()
        n = el.cos() * az.cos()
        u = el.sin()

        return [r * e, r * n, r * u]

    def horizontal_to_ecef(self, r, el, az):
        """ Convert r,el,az to ECEF """
        ## Convert to ENU
        e, n, u = Location.horizontal_to_enu(r, el, az)

        # Convert ENU to ECEF
        delta_ecef = self.get_ecef_delta_from_enu(e, n, u)
        my_ecef = np.array(self.get_ecef())

        return my_ecef + delta_ecef

    @staticmethod
    def GST(utc_date):
        """ Return the Greenwich Sidereal Time"""
        JD = tart_util.JulianDay(utc_date)
        D = JD - 2451545.0
        T = int(D // 36525)

        GMST = 18.697374558 + 24.06570982441908 * D

        # Now correct for precession of equioxes
        epsilon = 23.4393 - 0.0000004 * D
        L = 280.47 + 0.98565 * D
        omega = 125.04 - 0.052954 * D
        dphi = -0.000319 * math.sin(angle.deg_to_rad(omega)) - 0.000024 * math.sin(
            angle.deg_to_rad(2.0 * L)
        )
        eqeq = dphi * math.cos(angle.deg_to_rad(epsilon))

        return angle.from_hours(GMST + eqeq)

    @staticmethod
    def GHA(utc_date, ra):
        """ Calculate the Greenwich Hour Angle (GHA)"""
        return Location.GST(utc_date) - ra

    def LST(self, utc_date):
        """ Calculate the Local Sidereal Time """
        return Location.GST(utc_date) + self.lon

    def LHA(self, utc_date, ra):
        """ Calculate the Local Hour Angle (LHA)"""
        return self.LST(utc_date) - ra

    def horizontal_to_LHA(self, utc_date, el, az):
        phi_0 = self.lat.to_rad()

        tan_h = az.sin() / (az.cos() * math.sin(phi_0) + el.tan() * math.cos(phi_0))
        local_hour_angle = angle.atan(tan_h)
        return local_hour_angle

    def equatorial_to_horizontal(self, utc_date, ra, dec):
        """ Convert RA/Decl to Elevation Azimuth this location """
        # [Peter Duffett-Smith, Jonathan_Zwart] Practical Astronomy with calculator and spreadsheet
        lha = self.LHA(utc_date, ra)
        lat = self.lat

        el = angle.asin(dec.sin() * lat.sin() + dec.cos() * lat.cos() * lha.cos())

        az = angle.atan2(
            (-lha.sin()) * dec.cos(), (dec.sin() - lat.sin() * el.sin()) / lat.cos()
        )

        if az.to_degrees() < 0.0:
            az = angle.from_dms(360.0 + az.to_degrees())

        return el, az

    def horizontal_to_equatorial(self, utc_date, el, az):
        """Convert an azimuth and elevation to RA/Decl
        Useful for looking straight up, and working out the RA/Declination
        Return RA , Decl as angle objects
        """
        lat = self.lat
        dec_sin = (el.sin() * lat.sin()) + (el.cos() * lat.cos() * az.cos())
        dec = angle.asin(dec_sin)

        LST = self.LST(utc_date)

        H = angle.atan2(
            -1.0 * el.cos() * lat.cos() * az.sin(), (el.sin() - (lat.sin() * dec_sin))
        )

        ra = LST - H

        return ra.to_ra(), dec.to_declination()

    def horizontal_to_eci(self, r, el, az, utc_date):
        x, y, z = self.horizontal_to_ecef(r, el, az)
        xi, yi, zi = ecef_to_eci(utc_date, x, y, z)
        return [xi, yi, zi]

    def eci_to_horizontal(self, x, y, z, utc_date):
        xf, yf, zf = eci_to_ecef(utc_date, x, y, z)
        r, el, az = self.ecef_to_horizontal(xf, yf, zf)
        return [r, el, az]


"""Convenient helper function for the location of the Physics Department Roof"""
Dunedin = Location(
    lat=angle.from_dms(-45.86391200), lon=angle.from_dms(170.51348452), alt=46.5
)

"""Convenient helper function for the location of the rural TART"""
Dunedin_Farm = Location(
    lat=angle.from_dms(-45.851868), lon=angle.from_dms(170.545558), alt=266.5
)

"""Convenient helper function for somewhere cold and damp in the far north"""
Aachen = Location(lat=angle.from_dms(50.778), lon=angle.from_dms(6.086), alt=46.5)



#
# An object that represents a location in the Elevation and Azimuth, and
# handles projection into the l,m plane.
#
# Azimuth is measured clockwise from north (y - axis)
# Tim Molteno 2017-2019
#
import numpy as np

from tart.util.angle import wrap_180
from tart.imaging import imaging


class ElAz:
    def __init__(self, el, az):
        # Azimuth is measured fo
        self.el = el
        self.az = wrap_180(az)
        self.el_r = np.radians(self.el)
        self.az_r = np.radians(self.az)

        # Unit direction vector for the source un (u,v,w) where
        # u is east-west    u_hat = [1, 0, 0]
        # v is south-north  v_hat = [0, 1, 0]
        # w is straight up  w_hat = [0, 0, 1]
        # l,m are the cosines of the angle between the source vector and the
        # u and v axes respectively.
        self.l = -np.sin(self.az_r) * np.cos(self.el_r)
        self.m = np.cos(self.az_r) * np.cos(self.el_r)
        self.n = np.sin(self.el_r)
        self.s = np.array([self.l, self.m, self.n])

    def get_lm(self):
        return self.l, self.m

    def __repr__(self):
        return f"ElAz(el={self.el:5.2f}, az={self.az:5.2f})"

    '''
        Return the area of a unit-solid area of source dS
        in the l,m plane
    '''
    def get_lm_area(self, dS=1.0):
        return dS * np.sqrt(1.0 - self.l * self.l - self.m * self.m)

    def get_px(self, num_bins):
        ''' Get source location in pixels from it's direction
            cosines. These should be image coordinates in an
            image num_bins x num_bins essentially used as array
            indices.
        '''
        return imaging.get_lm_index(self.l, self.m, image_size=num_bins)

    @classmethod
    def from_pixel_indices(cls, x_pix, y_pix, num_bins):
        ''' Create from pixel coordinates. The center (0,0) is
            in the middle.
        '''
        # print(f"from_pixels({x_pix}, {y_pix})")
        n2 = num_bins // 2
        max_index = num_bins - 0.5

        x0 = n2
        # x_pix = np.floor(x0 - (m*max_index/2))
        # -2*(x_pix - x0)/max_index = m
        m = -2*(x_pix - x0)/max_index

        # y_px = np.floor(x0 + (l*max_index/2))
        # 2*(y_pix - x0) = l*max_index
        l = 2*(y_pix - x0) / max_index

        x = l*max_index/2
        y = m*max_index/2
        # print(f"centred_pixels({x}, {y})")
        r = np.sqrt(x*x + y*y)
        # print(f"r=({r}, {n2})")

        if r < n2:
            el_r = np.arccos(r/n2)
        else:
            raise ValueError(f"Source {x_pix}, {y_pix} is not in the sky.")
            el_r = 0.0
        # print(f"    el_r = {el_r}")
        atn = np.arctan2(y, x)
        # print(f"    atn = {np.degrees(atn)}")
        az_r = atn - np.radians(90)
        # print(f"    az_r = {az_r}")

        return ElAz(np.degrees(el_r), np.degrees(az_r))

    def get_px_window(self, num_bins, window_deg):
        ''' Get a pixel window around a source with width window_deg
            This is assumed to be an inverse FFT image with num_bins x num_bins
        '''
        x_i, y_i = self.get_px(num_bins)

        d = imaging.deg_to_pix(num_bins, window_deg)

        x_min = int(np.floor(x_i - d))
        x_max = int(np.ceil(x_i + d))
        y_min = int(np.floor(y_i - d))
        y_max = int(np.ceil(y_i + d))
        area = (x_max - x_min) * (y_max - y_min)
        return x_min, x_max, y_min, y_max, area

    def get_old_lm(self):
        R = 10 * np.pi * 2
        r = 90 - np.degrees(self.el_r)
        r = R * np.arctan(r / R)
        x = r * np.sin(self.az_r)
        y = r * np.cos(self.az_r)
        return [-x, y]


def from_json(source_json, el_limit_deg=0.0, jy_limit=1e5):
    src_list = []
    for src in source_json:
        try:
            if (src["el"] > el_limit_deg) and (src["jy"] > jy_limit):
                src_list.append(ElAz(src["el"], src["az"]))
        except:
            print(f"ERROR in catalog src={src}")

    return src_list


def get_source_coordinates(source_list):
    l_list = []
    m_list = []

    for src in source_list:
        l_list.append(src.l)
        m_list.append(src.m)

    return [l_list, m_list]



import json
import os

import h5py
import numpy as np

try:
    import cPickle as pickle
except:
    import pickle

from tart.operation import observation, settings
from tart.simulation.simulation_source import HorizontalSource
from tart.util import angle, utc


class Visibility:
    """
    A container class for visibilities from a single observation.
    """
    def __init__(self, config, timestamp,
                    phase_el, phase_az):
        self.phase_el = phase_el
        self.phase_az = phase_az
        self.config = config
        self.timestamp = timestamp

    @classmethod
    def from_obs(cls, obs, phase_el, phase_az):
        config = obs.config
        timestamp = obs.timestamp
        return cls(config, timestamp, phase_el, phase_az)

    @classmethod
    def from_config(cls, config, timestamp):
        phase_el = angle.from_dms(90.0)
        phase_az = angle.from_dms(0.0)
        return cls(config, timestamp, phase_el, phase_az)

    def set_visibilities(self, v, b):
        if not isinstance(b, list):
            raise RuntimeError(
                f"Baselines (type={type(b)}) must be passed as a list object"
            )
        self.baselines = b
        self.v = v

    r"""Rotated one, aimed at ra, decl

    Justification:

    Let
        s_1(t) = N(t) e^{j omega t}

    t_g^{01) = t_g^1 - t_g^0 (arrival at a1 - arrival at a0)
    if t_g^{01} < 0, then s_0 arrives later than s_1 (defined by antennas.get_geo_delay_horizontal)
    therefore:

        s_0(t) = s_1(t - t_g^{01})

    v(0,1) = <s_0(t) s_1^{*}(t)>
                 = <s_1(t - t_g^{01}) s_1^{*}(t)>
                 = <N(t) e^{j omega t} e^{-j omega t_g^{01}} N^*(t) e^{-j omega t}>
                 = e^{-j omega t_g^{01}} <N(t) N^*(t)>

    So, after rotation

     <N(t) N^*(t)> = v(0,1) e^{+j omega t_g^{01}}
    """

    def rotate(self, sky_location):
        from tart.simulation import antennas

        stopped_vis = []
        omega = self.config.get_operating_frequency() * 2.0 * np.pi
        # Now we must do fringe stopping

        loc = self.config.get_loc()
        el, az = loc.equatorial_to_horizontal(
            self.timestamp, sky_location.ra, sky_location.dec
        )
        hsource = HorizontalSource(r=9.0e99, azimuth=az, elevation=el)

        ant_pos = self.config.get_antenna_positions()
        for v, b in zip(self.v, self.baselines):
            a0 = antennas.Antenna(loc, ant_pos[b[0]])
            a1 = antennas.Antenna(loc, ant_pos[b[1]])

            tg = antennas.get_geo_delay_horizontal(a0, a1, hsource)
            # tg is t_a1 - t_a0
            # (negative if a1 is closer to source than a0)

            # print(b, omega*tg)
            v = v * np.exp(-1.0j * omega * tg)
            stopped_vis.append(v)

        self.phase_el = el
        self.phase_az = az
        self.v = stopped_vis

    def vis(self, i, j):
        if j == i:
            raise RuntimeError("Baseline [%d,%d] is invalid" % (i, j))
        if j < i:  # The first index should be before the second
            return np.conjugate(self.vis(j, i))
        for k, b in enumerate(self.baselines):
            if b == [i, j]:
                return self.v[k]
        raise RuntimeError("Baseline [%d,%d] is invalid" % (i, j))

    def get_closure_phase(self, i, j, k):
        return (
            np.angle(self.vis(i, j))
            + np.angle(self.vis(j, k))
            - np.angle(self.vis(i, k))
        )

    def toString(self):
        ret = ""
        for i, b in enumerate(self.baselines):
            ret += " V(%s)=%g, I%g" % (str(b), np.abs(self.v[i]), np.angle(self.v[i]))
        return ret

    def to_json(self):
        ret = {}
        vlist = []
        for i, b in enumerate(self.baselines):
            vlist.append({'i': b[0],
                          'j': b[1],
                          're': float(np.real(self.v[i])),
                          'im': float(np.imag(self.v[i]))})
        ret['data'] = vlist
        ret['timestamp'] = self.timestamp.isoformat()
        return ret

    def __repr__(self):
        return f"vis(ts={self.timestamp})"


def Visibility_Lsq(vis1, vis2):
    """ Return least square based on the phases of 2 visibilities """
    if vis1.config.get_num_antenna() == vis2.config.get_num_antenna():
        difflist = []
        for v1, v2 in zip(vis1.v, vis2.v):
            diff = np.abs(np.angle(v1) - np.angle(v2))
            if diff > np.pi:
                diff = 2.0 * np.pi - diff
            difflist.append(diff)
        diffarr = np.array(difflist)
        return np.power(diffarr, 2).sum()


def Visibility_From_Conf(config, timestamp, phase_el, phase_az):
    obs = observation.Observation(timestamp=timestamp, config=config)
    vis = Visibility.from_obs(obs, phase_el, phase_az)
    return vis


def Visibility_Save_JSON(vis, filename):
    json_data = {}
    json_data["timestamp"] = vis.timestamp
    json_data["phase_el"] = vis.phase_el.to_degrees()
    json_data["phase_az"] = vis.phase_az.to_degrees()
    json_data["config"] = vis.config.Dict
    json_data["baselines"] = vis.baselines
    json_data["vis"] = vis.v
    with open(filename, "w") as outfile:
        # The default=str handles datetime objects as strings
        json.dump(
            json_data,
            outfile,
            default=str,
            sort_keys=True,
            indent=4,
            separators=(",", ": "),
        )


"""
    Load and save lists of visibilities. Use filename extensions to deal with the deprecated .pkl files (or .vis files)
"""


def list_save(vis_list, ant_pos, cal_gain, cal_ph, filename):
    _, file_extension = os.path.splitext(filename)

    if file_extension == ".pkl" or file_extension == ".vis":
        to_pkl(vis_list, filename)
    elif file_extension == ".hdf":
        to_hdf5(vis_list, ant_pos, cal_gain, cal_ph, filename)
    else:
        err = f"Unknown visibility file extension {file_extension}"
        raise RuntimeError(err)


def list_load(filename):
    _, file_extension = os.path.splitext(filename)

    if file_extension == ".pkl":
        vis_list = from_pkl(filename)
        err_count = 0
        ret = []
        for v in vis_list:
            if isinstance(v, tuple):
                err_count += 1
            else:
                ret.append(v)
        if err_count > 0:
            print(

                    "Warning. Visibility file: %s had %i visibilities missing"
                    % (filename, err_count)

            )
        return ret
    elif file_extension == ".vis":
        vis_list = from_pkl(filename)
        return vis_list
    elif file_extension == ".hdf":
        vis_list = from_hdf5(filename)
        return vis_list
    else:
        raise RuntimeError(f"Unknown file extension {file_extension}")


"""
    Deal with Pickle files (Deprecated in favour of HDF5 files)
"""


def to_pkl(vis, filename):
    save_data = open(filename, "wb")
    pickle.dump(vis, save_data, pickle.HIGHEST_PROTOCOL)
    save_data.close()


def from_pkl(filename):
    with open(filename, "rb") as load_data:
        ret = pickle.load(load_data, encoding="latin1")
    return ret


"""
    Deal with HDF5 Files
"""


def to_hdf5(vis_list, ant_pos, cal_gain, cal_ph, filename):
    """Save the Observation object,
    to a portable HDF5 format
    """
    if not isinstance(vis_list, list):
        raise RuntimeError("vis_list must be a list of visibility objects")

    vis0 = vis_list[0]

    vis_data = [vis.v for vis in vis_list]
    vis_ts = [vis.timestamp.isoformat() for vis in vis_list]

    with h5py.File(filename, "w") as h5f:
        dt = h5py.special_dtype(vlen=str)
        conftype = h5py.special_dtype(vlen=bytes)

        conf_dset = h5f.create_dataset("config", (1,), dtype=conftype)
        conf_dset[0] = vis0.config.to_json()
        h5f.create_dataset(
            "phase_elaz", data=[vis0.phase_el.to_degrees(), vis0.phase_az.to_degrees()]
        )
        h5f.create_dataset("baselines", data=vis0.baselines)

        h5f.create_dataset("vis", data=np.array(vis_data, dtype=np.complex64))
        h5f.create_dataset("gains", data=np.array(cal_gain, dtype=np.float32))
        h5f.create_dataset("phases", data=np.array(cal_ph, dtype=np.float32))

        h5f.create_dataset("antenna_positions", data=np.array(ant_pos, dtype=np.float32))

        h5f.create_dataset("timestamp", data=np.array(vis_ts, dtype=object), dtype=dt)

        # ts_dset = h5f.create_dataset('timestamp', (len(vis_ts),), dtype=dt)
        # for i,ts in enumerate(vis_ts):
        # print(ts)
        # ts_dset[i] = ts


def from_hdf5(filename):
    """Load the Visibility object,
    from a portable HDF5 format
    """

    vis_list = []
    ret = {}

    with h5py.File(filename, "r") as h5f:
        config_json = np.bytes_(h5f["config"][0])
        config = settings.from_json(config_json)
        hdf_baselines = h5f["baselines"][:].tolist()
        ant_pos = h5f["antenna_positions"][:]
        hdf_phase_elaz = h5f["phase_elaz"][:]

        hdf_timestamps = h5f["timestamp"]
        timestamps = [utc.from_string(x) for x in hdf_timestamps]

        hdf_vis = h5f["vis"][:]
        config.set_antenna_positions(ant_pos)
        for ts, v in zip(timestamps, hdf_vis):
            vis = Visibility.from_config(config=config, timestamp=ts)
            vis.set_visibilities(v=v, b=hdf_baselines)
            vis.phase_el = hdf_phase_elaz[0]
            vis.phase_az = hdf_phase_elaz[1]
            vis_list.append(vis)

        ret["vis_list"] = vis_list
        ret["config_json"] = config_json
        ret["config"] = config
        ret["ant_pos"] = ant_pos
        ret["timestamps"] = timestamps
        ret["baselines"] = hdf_baselines
        ret["gain"] = h5f["gains"][:]
        ret["phase_offset"] = h5f["phases"][:]

    return ret




import numpy as np

from tart.operation import settings
from tart.simulation import antennas, radio, simulation_source

if __name__ == "__main__":
    import numpy as np

    rad = radio.Max2769B(sample_duration=1e-5)
    sources = [
        simulation_source.SimulationSource(
            amplitude=1.0,
            elevation=30.0,
            azimuth=0.0,
            sample_duration=rad.sample_duration,
        )
    ]

    print("Radio Sampling Rate %g" % rad.sampling_rate)

    config = settings.Settings("../../tools/operation/telescope_config.json")
    ants = [antennas.Antenna(config.get_loc(), pos) for pos in config.ant_positions]
    ant_sigs = antennas.antennas_new(ants, sources, rad.timebase)

    num_radio_samples = (len(rad.timebase) / rad.freq_mult) + 1
    sampled_signals = np.zeros((config.get_num_antenna(), num_radio_samples))
    for i in range(config.get_num_antenna()):
        sampled_signals[i, :] = rad.sampled_signal(ant_sigs[i, :])

    print("The (truncated) array of radio signals is; ", sampled_signals[:, :10])
    print("Signal from each radio are in rows")



import numpy as np

try:
    from pylab import plot, show, subplot, xlabel, ylabel

    PYLAB_AVAILABLE = True
except ImportError:
    PYLAB_AVAILABLE = False

import numpy.fft as fft


def getSpectrum(y, Fs):
    """
    Get a Single-Sided Amplitude Spectrum of y(t)
    """

    n = len(y)  # length of the signal
    k = np.arange(n)
    T = n / Fs
    frq = k / T  # two sides frequency range
    frq = frq[list(range(n // 2))]  # one side frequency range

    Y = fft.fft(y) / n  # fft computing and normalization
    Y = Y[list(range(n // 2))]

    return (frq, np.array(abs(Y)))


def plotSpectrum(y, Fs, c="red", label="powerspectrum", doPlot=False):
    """
    Plots a Single-Sided Amplitude Spectrum of y(t)
    """
    if not PYLAB_AVAILABLE:
        raise ImportError(
            "matplotlib/pylab is required for plotSpectrum. Install with: pip install matplotlib"
        )

    frq, absY = getSpectrum(y, Fs)

    plot(frq, absY, c=c, label=label)  # plotting the spectrum
    xlabel("Freq (Hz)")
    ylabel("|Y(freq)|")
    return (frq, absY)


if __name__ == "__main__":
    if not PYLAB_AVAILABLE:
        print("matplotlib/pylab is required for plotting examples")
        exit(1)

    Fs = 150.0
    # sampling rate
    Ts = 1.0 / Fs
    # sampling interval
    t = np.arange(0, 1, Ts)  # time vector

    ff = 5
    # frequency of the signal
    y = np.sin(2 * np.pi * ff * t)

    subplot(2, 1, 1)
    plot(t, y)
    xlabel("Time")
    ylabel("Amplitude")
    subplot(2, 1, 2)
    plotSpectrum(y, Fs)
    show()



import unittest

import numpy as np

from tart.imaging import visibility
from tart.operation import settings
from tart.util import angle, skyloc, utc

VIS_DATA_FILE = "tart/test/test_data/fpga_2019-02-22_05_11_41.765212.vis"
ANT_POS_FILE = "tart/test/test_calibrated_antenna_positions.json"


def dummy_vis():
    config = settings.from_file("tart/test/test_telescope_config.json")
    ret = visibility.Visibility.from_config(config=config, timestamp=utc.now())
    b = []
    v = []
    for j in range(24):
        for i in range(24):
            if i > j:
                b.append([j, i])
                v.append(np.random.uniform(0, 1) + np.random.uniform(0, 1) * (1.0j))
    ret.set_visibilities(b=b, v=np.array(v))
    return ret


def dummy_vis_list():
    ret = []
    for i in range(2):
        ret.append(dummy_vis())
    return ret


class TestVisibility(unittest.TestCase):
    def setUp(self):
        self.v_array = dummy_vis_list()
        self.v_array[0].config.load_antenna_positions(
            cal_ant_positions_file=ANT_POS_FILE
        )
        self.ant_pos=self.v_array[0].config.get_antenna_positions()

    def check_vis(self, dut, dut2):
        delta = np.sum(np.abs(np.array(dut.v) - np.array(dut2.v)))
        self.assertTrue(delta <= 1.0e10)
        self.assertTrue((np.array(dut.baselines) == np.array(dut2.baselines)).all())
        self.assertEqual(dut.timestamp, dut2.timestamp)

        keys = dut.config.Dict.keys()
        for k in keys:
            self.assertEqual(dut.config.Dict[k], dut2.config.Dict[k])

    def test_load_save(self):
        dut = dummy_vis()
        visibility.Visibility_Save_JSON(self.v_array[0], "test_vis.json")
        # TODO add tests here

    def test_list_load_save_hdf(self):
        dut_list = dummy_vis_list()
        fname = "test_vis_list_io.hdf"
        visibility.list_save(dut_list, ant_pos=self.ant_pos,
                             cal_gain=[0], cal_ph=[0],
                             filename=fname)  # ant_pos, cal_gain, cal_ph, filename
        dut2_list = visibility.list_load(fname)['vis_list']
        for x, y in zip(dut_list, dut2_list):
            self.check_vis(x, y)


    def test_hdf5(self):
        dut = dummy_vis()
        fname = "test_vis.hdf"
        visibility.to_hdf5([dut],  ant_pos=self.ant_pos,
                           cal_gain=[0], cal_ph=[0],
                           filename=fname)
        dut2 = visibility.from_hdf5(fname)
        print(dut2)
        vl = dut2['vis_list']
        self.check_vis(dut, vl[0])


    def test_zero_rotation(self):
        v = self.v_array[0]
        v_before = np.array(v.v)  # Copy the visibilities
        el1 = v.phase_el
        az1 = v.phase_az

        ra, decl = v.config.get_loc().horizontal_to_equatorial(v.timestamp, el1, az1)

        v.rotate(skyloc.Skyloc(ra, decl))

        v_after = np.array(v.v)
        self.assertEqual(v.phase_el.to_degrees(), el1.to_degrees())

        for x, y in zip(v_before, v_after):
            self.assertAlmostEqual(np.abs(x), np.abs(y), 4)

    def test_full_rotation(self):
        v = self.v_array[0]
        v_before = np.array(v.v)  # Copy the visibilities
        el1 = v.phase_el
        az1 = v.phase_az
        ra, decl = v.config.get_loc().horizontal_to_equatorial(v.timestamp, el1, az1)

        # Pick a baseline

        # Calculate the angle to offset the RA, DEC to produce a full rotation
        #

        v.rotate(skyloc.Skyloc(ra + angle.from_dms(0.05), decl))
        v_after = np.array(v.v)
        self.assertAlmostEqual(v.phase_el.to_degrees(), el1.to_degrees(), 0)
        for x, y in zip(v_before, v_after):
            self.assertAlmostEqual(np.angle(x), np.angle(y), 1)
            self.assertAlmostEqual(np.abs(x), np.abs(y), 2)



import datetime
import unittest

from astropy import time

from tart.util import utc


class TestUTC(unittest.TestCase):

    def compare(self, utcd, utcd2):
        self.assertEqual(utcd2.year, utcd.year)
        self.assertEqual(utcd2.month, utcd.month)
        self.assertEqual(utcd2.day, utcd.day)
        self.assertEqual(utcd2.hour, utcd.hour)
        self.assertEqual(utcd2.minute, utcd.minute)
        self.assertEqual(utcd2.second, utcd.second)
        self.assertAlmostEqual((utcd2 - utcd).total_seconds(), 0.0, 3)


    def test_basic(self):
        utcd = utc.utc_datetime(year=2009, month=7, day=22, hour=5, minute=9, second=50.3)
        apyt = time.Time('2009-07-22 05:09:50.3', format='iso', scale='utc')
        apyt2 = time.Time(utcd, scale='utc')

        utcd2 = apyt.to_datetime(utc.UTC)
        obstime = apyt2.to_datetime(utc.UTC)

        self.compare(utcd2, utcd)

        self.assertEqual((utcd - obstime).total_seconds(), 0.0)


    def test_now(self):
        self.assertRaises(RuntimeError, utc.to_utc, datetime.datetime.now())

    def test_parsing(self):
        now = utc.now()

        # As used by create_direct_vis_dict
        rep = utc.to_string(now)

        # As used by observation.from_hdf5
        timestamp = utc.from_string(rep)

        self.compare(timestamp, now)
        self.assertEqual(timestamp.isoformat(), now.isoformat())



## Utility functions for imaging
import numpy as np

from tart.imaging import synthesis
from tart.operation import settings

# Following for simulation only
from tart.util import angle
from tart.imaging import antenna_model
from tart.imaging import radio_source
from tart.imaging import location
from tart.simulation import antennas
from tart.simulation import radio
from tart.simulation import skymodel
from tart.imaging import calibration


def deg_to_pix(num_bins, deg):
    pix_per_degree = num_bins / 180.0
    d = deg*pix_per_degree
    return d


def get_lm_index(l, m, image_size):
    ''' The l axis is along the +x in the image plane.
        The numpy array index for this is the second index. For example
        the point [image_size , 0] is in the LOWER LEFT of the image when displayed
        by imshow.

        https://matplotlib.org/stable/users/explain/artists/imshow_extent.html#imshow-extent

         0, 0  -> [image_size // 2, image_size // 2]
        -1, 1  -> [0 , 0]
        -1, -1 -> [image_size-1, 0]
        1, -1  -> [image_size-1, image_size-1]
        1,  1  -> [0, image_size-1]
    '''
    x0 = image_size // 2
    max_index = image_size - 0.5
    index0 = np.floor(x0 - (m*max_index/2))
    index1 = np.floor(x0 + (l*max_index/2))
    return int(index0), int(index1)


def get_baseline_indices(num_ant):
    bl_indices = []
    for i in range(num_ant-1):
        for j in range(i + 1, num_ant):
            bl_indices.append([i, j])
    return bl_indices


def get_baselines(ant_pos):
    num_ant = ant_pos.shape[0]
    bl_indices = get_baseline_indices(num_ant)
    i_indices = bl_indices[:, 0]
    j_indices = bl_indices[:, 1]

    return ant_pos[j_indices, :] - ant_pos[i_indices, :]


def ant_pos_to_uv(ant_pos, i, j):
    return ant_pos[j] - ant_pos[i]


def apply_complex_gains(v_complex, gains_complex, i, j):
    return v_complex * gains_complex[i] * np.conj(gains_complex[j])


def ifft_imaging(uv_plane, module=np.fft):
    return module.fft.fftshift(
        module.fft.ifft2(
            module.fft.ifftshift(uv_plane)
            )
        )


def uv_index(u, v, num_bins, uv_max):
    ''' A little function to produce the index into the u-v array
        for a given value (u, measured in wavelengths)
    '''
    middle = num_bins // 2

    u_pix = middle + (u / uv_max)*(num_bins/2)
    v_pix = middle + (v / uv_max)*(num_bins/2)

    return int(u_pix), int(v_pix)


def grid_visibility(uv_plane, v_complex, baselines):
    num_bins = uv_plane.shape[0]
    uv_max = num_bins / 4   # (1.2 * np.pi)

    n_vis = len(v_complex)
    for i in range(n_vis):
        v = v_complex[i]
        uu, vv, ww = baselines[i]

        u_idx, v_idx = uv_index(uu, vv, num_bins, uv_max)
        uv_plane[u_idx, v_idx] += v

        # Place the conjugate visibility at -uu, -vv
        u_idx, v_idx = uv_index(--uu, --vv, num_bins, uv_max)
        uv_plane[u_idx, v_idx] += np.conj(v)

    return uv_max


def rotate_vis(rot_degrees, cv, reference_positions):
    '''
        Note. This rotates counter_clockwise
        (antennas in the north move towards the east)
    '''
    conf = cv.vis.config

    new_positions = settings.rotate_location(
        rot_degrees, np.array(reference_positions).T
    )
    conf.set_antenna_positions((np.array(new_positions).T).tolist())


def image_from_calibrated_vis(cv, nw, num_bin):
    cal_syn = synthesis.Synthesis_Imaging([cv])

    cal_ift, cal_extent = cal_syn.get_ift(nw=nw, num_bin=num_bin)
    # beam = cal_syn.get_beam(nw=nw, num_bin=num_bin, use_kernel=False)
    n_fft = len(cal_ift)
    assert n_fft == num_bin

    bin_width = (max(cal_extent) - min(cal_extent)) / float(n_fft)

    return cal_ift, cal_extent, n_fft, bin_width


def get_clock_hands(timestamp):
    # ############## HOUR HAND ###########################
    #
    # The pattern rotates once every 12 hours
    #
    hour_azimuth = timestamp.hour*30.0 + timestamp.minute/2.0

    hour_sources = [{'el': el, 'az': -hour_azimuth} for el in [85, 75, 65, 55]]

    # ############## MINUTE HAND ###########################
    #
    # The pattern rotates 360 deg once every 1 hour
    #
    minute_azimuth = timestamp.minute*6.0 + timestamp.second/10.0

    minute_sources = [{'el': el, 'az': -minute_azimuth} for el in [90, 80, 70, 60, 50, 40, 30]]

    return hour_sources, minute_sources


# Helper function for testbenches
def get_clock_vis(config, timestamp):

    loc = location.Location(angle.from_dms(config.get_lat()),
                            angle.from_dms(config.get_lon()),
                            config.get_alt())

    ant_pos = config.get_antenna_positions()
    num_ant = len(ant_pos)

    ANTS = [antennas.Antenna(loc, enu=pos) for pos in ant_pos]
    ANT_MODELS = [antenna_model.GpsPatchAntenna() for i in range(num_ant)]
    RAD = radio.Max2769B(n_samples=2**12, noise_level=np.zeros(num_ant))

    hour_sources, minute_sources = get_clock_hands(timestamp)

    sim_sky = skymodel.Skymodel(0, location=loc, gps=0,
                                thesun=0, known_cosmic=0)

    for m in hour_sources + minute_sources:
        src = radio_source.ArtificialSource(loc, timestamp, r=1e6,
                                            el=m['el'], az=m['az'])
        sim_sky.add_src(src)

    sources = sim_sky.gen_photons_per_src(timestamp, radio=RAD,
                                          config=config, n_samp=1)

    v_sim = antennas.antennas_simp_vis(
        ANTS, ANT_MODELS, sources, timestamp, config, RAD.noise_level
    )

    cv = calibration.CalibratedVisibility(v_sim)

    return cv, hour_sources, minute_sources






import time

import numpy as np
import pyfftw


def hilbert(s, debug=False):
    '''
    http://au.mathworks.com/help/signal/ref/hilbert.html
    The analytic signal for a sequence x has a one-sided Fourier transform. That is, the transform vanishes for negative frequencies. To approximate the analytic signal, hilbert calculates the FFT of the input sequence, replaces those FFT coefficients that correspond to negative frequencies with zeros, and calculates the inverse FFT of the result.

    In detail, hilbert uses a four-step algorithm:

    It calculates the FFT of the input sequence, storing the result in a vector x.
    It creates a vector h whose elements h(i) have the values:
    1 for i = 1, (n/2)+1
    2 for i = 2, 3, ... , (n/2)
    0 for i = (n/2)+2, ... , n
    It calculates the element-wise product of x and h.
    It calculates the inverse FFT of the sequence obtained in step 3 and returns the first n elements of the result.
    This algorithm was first introduced in [2]. The technique assumes that the input signal, x, is a finite block of data. This assumption allows the function to remove the spectral redundancy in x exactly. Methods based on FIR filtering can only approximate the analytic signal, but they have the advantage that they operate continuously on the data. See Single-Sideband Amplitude Modulation for another example of a Hilbert transform computed with an FIR filter.'''

    s_0 = time.time()
    S = np.fft.fft(s)
    if debug:
        print('fft', time.time()-s_0)
    s_1 = time.time()

    n = len(s)
    h = np.zeros(n)
    h[0] = 1.
    h[n/2] = 1.
    h[1:n/2] = 2.
    #h[n/2+1:] = 0.
    if debug:
        print('setup', time.time()-s_1)
    s_2 = time.time()
    ret = np.fft.ifft(h*S)[:n].imag
    if debug:
        print('ifft', time.time()-s_2)

    return ret


def hilbert_fftw(s, debug=False, dtype='complex64'):
    ''' fftw drop replacement for scipy_fftpack.hilbert
    beware of sign of returned seq. is '-'

    http://au.mathworks.com/help/signal/ref/hilbert.html
    The analytic signal for a sequence x has a one-sided Fourier transform. That is, 
    the transform vanishes for negative frequencies. To approximate the analytic signal, 
    hilbert calculates the FFT of the input sequence, replaces those FFT coefficients 
    that correspond to negative frequencies with zeros, and calculates the inverse FFT of the result.

    In detail, hilbert uses a four-step algorithm:

    It calculates the FFT of the input sequence, storing the result in a vector x.
    It creates a vector h whose elements h(i) have the values:
    1 for i = 1, (n/2)+1
    2 for i = 2, 3, ... , (n/2)
    0 for i = (n/2)+2, ... , n
    It calculates the element-wise product of x and h.
    It calculates the inverse FFT of the sequence obtained in step 3 and returns the first n elements of the result.
    This algorithm was first introduced in [2]. The technique assumes that the input signal, x, 
    is a finite block of data. This assumption allows the function to remove the spectral redundancy in x exactly. 
    Methods based on FIR filtering can only approximate the analytic signal, but they have the advantage that 
    they operate continuously on the data. See Single-Sideband Amplitude Modulation for another example of a 
    Hilbert transform computed with an FIR filter.
    '''

    n = len(s)
    pyfftw.interfaces.cache.enable()
    pyfftw.interfaces.cache.set_keepalive_time(50.0)
    align = pyfftw.simd_alignment

    write_wisdom = False
    try:
        wisdom = pickle.load(open("wisdom_hilbert.wis", "rb"))
        pyfftw.import_wisdom(wisdom)
    except:
        write_wisdom = True
        print('no wisdom file')

    fft_in = pyfftw.empty_aligned(n, dtype=dtype, n=align)
    fft_out = pyfftw.empty_aligned(n, dtype=dtype, n=align)
    ifft_in = pyfftw.empty_aligned(n, dtype=dtype, n=align)
    ifft_out = pyfftw.empty_aligned(n, dtype=dtype, n=align)
    fft_machine = pyfftw.FFTW(fft_in, fft_out, direction='FFTW_FORWARD', flags=('FFTW_ESTIMATE',), threads=8)
    ifft_machine = pyfftw.FFTW(ifft_in, ifft_out, direction='FFTW_BACKWARD', flags=('FFTW_ESTIMATE',), threads=8)

    if write_wisdom:
        wisdom = pyfftw.export_wisdom()
        pickle.dump(wisdom, open("wisdom_hilbert.wis", "wb"))

    s_0 = time.time()
    fft_in[:] = s
    S = fft_machine()
    if debug:
        print('fft', time.time()-s_0)

    s_1 = time.time()
    h = np.zeros(n)
    h[0] = 1.
    h[n/2] = 1.
    h[1:n/2] = 2.
    if debug:
        print('setup', time.time()-s_1)
    s_2 = time.time()
    ifft_in[:] = h*S
    ret = ifft_machine()
    if debug:
        print('ifft', time.time()-s_2)
    return -ret[:n].imag


if __name__ == '__main__':
    from pyfftw.interfaces.scipy_fftpack import hilbert as scipy_fftpack_hilbert
    x = np.random.randint(0, 2, 2**21)*2-1
    start = time.time()
    a = -hilbert(x)
    print('hilbert took', time.time()-start)
    start = time.time()
    b = hilbert_fftw(x, debug=False)
    print('max took', time.time()-start)

    start = time.time()
    c = scipy_fftpack_hilbert(x)
    print('hilbert fftw took', time.time()-start)

    a = np.sign(a)
    b = np.sign(b)
    c = np.sign(c)
    print(np.allclose(a, b))
    print(np.allclose(b, c))
    print(np.allclose(a, c))



import unittest

import matplotlib.pyplot as plt
import numpy as np

from tart.imaging import antenna_model
from tart.operation import settings
from tart.simulation import antennas, simulation_source, spectrum
from tart.simulation.radio import *
from tart.util import angle, utc


class TestMax2769B(unittest.TestCase):
    def setUp(self):
        self.config = settings.from_file("./tart/test/test_telescope_config.json")
        self.config.load_antenna_positions(
            cal_ant_positions_file="./tart/test/test_calibrated_antenna_positions.json"
        )
        # noiselvls =  0.1.*np.ones(config.get_num_antenna())
        num_ant = self.config.get_num_antenna()
        noiselvls = 0.0 * np.ones(num_ant)
        self.rad = Max2769B(noise_level=noiselvls)
        self.sources = [
            simulation_source.SimulationSource(
                r=1e9,
                amplitude=1.0,
                azimuth=angle.from_dms(0.0),
                elevation=angle.from_dms(90.0),
                sample_duration=self.rad.sample_duration,
            )
        ]
        self.ants = [
            antennas.Antenna(self.config.get_loc(), pos)
            for pos in self.config.get_antenna_positions()
        ]
        self.ant_models = [antenna_model.GpsPatchAntenna() for i in range(num_ant)]
        self.utc_date = utc.now()

    def test_get_obs(self):

        plt.figure()
        timebase = np.arange(0, self.rad.sample_duration, 1.0 / self.rad.sampling_rate)
        ant_sigs = antennas.antennas_signal(
            self.ants, self.ant_models, self.sources, timebase
        )
        timebase = np.arange(
            0, self.rad.sample_duration, 1.0 / self.rad.sampling_rate
        )
        rad_sig_full = self.rad.sampled_signal(ant_sigs[0, :], 0, sample_duration=self.rad.sample_duration)
        obs_full = self.rad.get_full_obs(ant_sigs, self.utc_date, self.config, timebase)

        ant_sigs_simp = antennas.antennas_simplified_signal(
            self.ants,
            self.ant_models,
            self.sources,
            self.rad.baseband_timebase,
            self.rad.int_freq,
        )
        obs_simp = self.rad.get_simplified_obs(
            ant_sigs_simp, self.utc_date, self.config
        )

        if True:
            freqs, spec_full_before_obs = spectrum.plotSpectrum(
                rad_sig_full, self.rad.ref_freq, label="before", c="green"
            )
            freqs, spec_full = spectrum.plotSpectrum(
                obs_full.get_antenna(1), self.rad.ref_freq, label="full", c="blue"
            )
            freqs, spec_simp = spectrum.plotSpectrum(
                obs_simp.get_antenna(1), self.rad.ref_freq, label="simp", c="red"
            )
            plt.legend()
            plt.show()
        self.assertTrue((spec_full_before_obs == spec_full).all(), True)

        #plt.figure()
        #plt.plot(freqs, (spec_simp - spec_full) / spec_full)
        # plt.show()

        print(len(obs_full.get_antenna(1)), obs_full.get_antenna(1).mean())
        print(len(obs_simp.get_antenna(1)), obs_simp.get_antenna(1).mean())

        # self.assertEqual(self.settings.get_num_antenna(), 5)
        # self.assertLess(d.std(), 0.1)


if __name__ == "__main__":

    a = TestMax2769B()
    a.test_get_obs()



# A UTC class that allows the construction of datetime objects with the timezone of UTC.
# Tim Molteno 2013-2022
#
from datetime import datetime, timezone

from dateutil import parser

UTC = timezone.utc


def utc_datetime(year, month, day, hour=0, minute=0, second=0.0):
    s = int(second)
    us = int((second - int(second)) * 1000000)
    return datetime(year=year, month=month, day=day,
                    hour=hour, minute=minute, second=s,
                    microsecond=us, tzinfo=timezone.utc)


def to_utc(dt):
    if dt.tzinfo is None:
        err = f"Attempting to convert non-utc timezone ({dt}) to utc: {dt.tzinfo}"
        raise RuntimeError(err)
    if dt.tzinfo.utcoffset(dt).total_seconds() != 0:
        err = f"Attempting to convert non-utc timezone ({dt}) to utc: {dt.tzinfo}"
        raise RuntimeError(err)
    return dt


def now():
    return datetime.now(timezone.utc)


def from_string(repr):
    return to_utc(parser.parse(repr))


def to_string(timestamp):
    return timestamp.isoformat()





def telescope_signals(ref_freq, freq_mult, sample_duration, baseline, angle_1, angle_2):

    import numpy as np

    SNR = 20  # Signal to noise ratio

    # Introduce the geometric delay
    delay_1 = baseline * np.sin(angle_1) / 2.99793e8
    delay_2 = baseline * np.sin(angle_2) / 2.99793e8

    # how many source rates is this delay
    sample_delay_1 = int(delay_1 * (ref_freq * freq_mult))
    sample_delay_2 = int(delay_2 * (ref_freq * freq_mult))

    print("delay_1 = ", delay_1)
    print("delay_2 = ", delay_2)
    print("sample_delay_1 = ", sample_delay_1)
    print("sample_delay_2 = ", sample_delay_2)

    # Calculate the number of points required
    n = int(ref_freq * freq_mult * sample_duration)

    # create two signals
    source_1 = np.random.random(n + sample_delay_1)  # Add gaussian noise
    source_2 = np.random.random(n + sample_delay_2)  # Add gaussian noise
    noise_1 = np.random.random(n) / SNR  # Add uncorrelated gaussian noise
    noise_2 = np.random.random(n) / SNR  # Add uncorrelated gaussian noise
    s1 = source_1[
        :n
    ]  # + source_2[:n] + noise_1 # Un-delayed signal with instrument noise
    s2 = source_1[
        sample_delay_1 : n + sample_delay_1
    ]  # + source_2[sample_delay_2:n+sample_delay_2]+ noise_2 # Delayed signal with instrument noise

    return (s1, s2)


if __name__ == "__main__":
    import numpy as np

    ref_freq = 16.368e6
    freq_mult = 256
    sample_duration = 1e-3
    ref_div = 16
    main_div = 1536
    bandwidth = 2.5e6
    order = 5
    baseline = 20.0  # Meters
    angle_1 = 30 * np.pi / 180  # Radians
    angle_2 = 60 * np.pi / 180  # Radians

    telescope_signals(
        ref_freq,
        freq_mult,
        sample_duration,
        ref_div,
        main_div,
        bandwidth,
        order,
        baseline,
        angle_1,
        angle_2,
    )



import datetime
import unittest

import numpy as np

from tart.imaging import gps_time
from tart.imaging.ephemerides_proxy import EphemeridesProxy
from tart.imaging.ephemeris import Ephemeris
from tart.util import utc


class TestEphemeris(unittest.TestCase):
    def setUp(self):
        self.ep = EphemeridesProxy.Instance()
        self.s = self.ep.server
        self.t = utc.utc_datetime(2002, 10, 31, 0, 2, 2)

        self.eph_hash = self.s.get_ephemeris(self.t.isoformat(), 1)

    def test_eph(self):
        self.assertEqual(self.eph_hash["toe"], 345600.0)
        self.assertEqual(self.eph_hash["roota"], 5153.57448578)

        eph = Ephemeris(self.eph_hash)
        self.assertEqual(self.eph_hash["toe"], eph.toe)

        sow = 123.0
        pos = eph.get_location(sow)

        gpst = gps_time.GpsTime.from_time(self.t)
        pos2 = eph.get_sv_position(gpst)

        t = self.t
        for i in range(100):
            t = t + datetime.timedelta(seconds=200.0)
            gpst = gps_time.GpsTime.from_time(t)
            pos = eph.get_sv_position(gpst)
            pos_remote = self.s.get_sv_position(t.isoformat(), 1)
            diff = np.array(pos) - np.array(pos_remote)
            dr = np.sqrt(diff.dot(diff))
            dt = abs(eph.toe - gpst.sow())
            if dt > 3600:
                self.eph_hash = self.s.get_ephemeris(t.isoformat(), 1)
                eph = Ephemeris(self.eph_hash)

            self.assertLess(dr, 3.0)  # Maximum difference of 3 meters
